from Cryptodome.Cipher import AES
import gzip
from secp256k1 import PrivateKey, PublicKey
from .constants import bip340tag, preamble, DATA_LENGTH
from typing import Iterable, Tuple, Any


def compress(pairs: Iterable[Tuple[str, str]]) -> bytes:
    """Compress the pairs, padding with zeroes to DATA_LENGTH, raising an exception if the result is too large"""
    raw = bytes()
    for title, contents in pairs:
        raw += bytes(title, encoding="utf8")
        raw += bytes(1)
        raw += bytes(contents, encoding="utf8")
        raw += bytes(1)
    ret = gzip.compress(raw)
    if len(ret) > DATA_LENGTH:
        raise ValueError("Compressed length too great!")

    return ret.ljust(DATA_LENGTH, bytes(1))


def aes(aeskey: bytes, compressed: bytes) -> bytes:
    """Encrypt the compressed data using the given key"""
    assert len(aeskey) == 32
    assert len(compressed) == DATA_LENGTH
    encrypter = AES.new(key=aeskey, mode=AES.MODE_CTR, nonce=bytes(8))

    return encrypter.encrypt(compressed)


def get_aeskey(privkey: PrivateKey, pubkey: PublicKey) -> bytes:
    return pubkey.ecdh(privkey.private_key)


def contents(writer: PublicKey, reader: PublicKey, gen: int, aes: bytes) -> bytes:
    return writer.serialize() + reader.serialize() + gen.to_bytes(8, "big") + aes


def sign(writer: PrivateKey, contents: bytes) -> bytes:
    return writer.schnorr_sign(contents, bip340tag)


def encode(secretkey: PrivateKey, readerpubkey: PublicKey, generation: int, *pairs: Any) -> bytes:
    comp = compress(pairs)
    enc = aes(get_aeskey(secretkey, readerpubkey), comp)
    cont = contents(secretkey.pubkey, readerpubkey, generation, enc)
    return preamble + sign(secretkey, cont) + cont



from Cryptodome.Cipher import AES
import gzip
from secp256k1 import PrivateKey, PublicKey
from .constants import bip340tag, preamble, DATA_LENGTH, FULL_LENGTH
from .encode import get_aeskey
from typing import Tuple, List, Optional


def decompress(comp: bytes) -> Optional[List[Tuple[str, str]]]:
    """Compress into pairs"""
    uncomp = gzip.decompress(comp)
    # Split by 0 byte
    fields = uncomp.split(sep=bytes(1))
    # That gives us a final empty field, which we ignore...
    if len(fields) != 0 and len(fields) % 2 != 1:
        return None

    ret = []
    for i in range(0, len(fields) - 1, 2):
        ret.append((fields[i].decode('utf8'), fields[i + 1].decode('utf8')))
    return ret


def unaes(aeskey: bytes, encrypted: bytes) -> bytes:
    """Decrypt the compressed data using the given key"""
    assert len(aeskey) == 32
    assert len(encrypted) == DATA_LENGTH
    decrypter = AES.new(key=aeskey, mode=AES.MODE_CTR, nonce=bytes(8))

    return decrypter.decrypt(encrypted)


def split_parts(after_preamble: bytes) -> Tuple[bytes, PublicKey, PublicKey, int, bytes]:
    """Split into sig, writer, reader, gen, aes"""
    try:
        wkey = PublicKey(after_preamble[64:64 + 33], raw=True)
    except Exception:
        # FIXME: secp256k1 should use a decent exception here!
        raise ValueError("Invalid wkey {}".format(after_preamble[64:64 + 33].hex()))

    try:
        rkey = PublicKey(after_preamble[64 + 33:64 + 33 + 33], raw=True)
    except Exception:
        # FIXME: secp256k1 should use a decent exception here!
        raise ValueError("Invalid rkey {}".format(after_preamble[64 + 33:64 + 33 + 33].hex()))
    return (after_preamble[0:64], wkey, rkey,
            int.from_bytes(after_preamble[64 + 33 + 33:64 + 33 + 33 + 8], "big"),
            after_preamble[64 + 33 + 33 + 8:])


def check_sig(after_preamble: bytes) -> bool:
    assert len(after_preamble) == FULL_LENGTH

    try:
        sig, wkey, _, _, _ = split_parts(after_preamble)
    except ValueError:
        return False

    return wkey.schnorr_verify(after_preamble[64:], sig, bip340tag)


def deconstruct(cmetadata: bytes) -> Tuple[PublicKey, PublicKey, int, bytes]:
    """Deconstructs a cmetadata into reader, writer, generation and post-preamble"""
    if not cmetadata.startswith(preamble):
        raise ValueError("Incorrect preamble")
    after_preamble = cmetadata[len(preamble):]
    if len(after_preamble) != FULL_LENGTH:
        raise ValueError("Expected {} bytes after preamble, got {}"
                         .format(FULL_LENGTH, len(after_preamble)))

    _, wkey, rkey, gen, _ = split_parts(after_preamble)
    return wkey, rkey, gen, after_preamble


def decode(secretkey: PrivateKey, cmetadata: bytes) -> Optional[List[Tuple[str, str]]]:
    if not cmetadata.startswith(preamble):
        return None
    after_preamble = cmetadata[len(preamble):]
    if len(after_preamble) != FULL_LENGTH:
        return None
    if not check_sig(after_preamble):
        return None
    sig, wkey, rkey, gen, aes = split_parts(after_preamble)
    # FIXME: secp256k1 != doesn't work like you expect!
    if rkey.serialize() != secretkey.pubkey.serialize():
        return None

    comp = unaes(get_aeskey(secretkey, wkey), aes)
    return decompress(comp)



"""centurymetadata: routines to handle long-persistent, small encrypted data.

"""
from .constants import verheader, preamble, DATA_LENGTH, FULL_LENGTH, RECORD_LENGTH
from .encode import compress, aes, get_aeskey, contents, sign, encode
from .decode import decompress, unaes, check_sig, decode, deconstruct

__all__ = [
    "compress",
    "aes",
    "get_aeskey",
    "contents",
    "sign",
    "encode",
    "DATA_LENGTH",
    "FULL_LENGTH",
    "RECORD_LENGTH",
    "verheader",
    "preamble",
    "decompress",
    "unaes",
    "check_sig",
    "decode",
    "deconstruct",
]



verheader = b"centurymetadata v0\0"
preamble = verheader + b"""SIG[64]|WRITER[33]|READER[33]|GEN[8]|AES[8054]

SIG: BIP-340 SHA256(TAG|TAG|WRITER|READER|GEN|AES)
WRITER, READER: secp256k1 x-only keys
TAG: SHA256("centurymetadata v0"[18])
AESKEY: SHA256(EC Diffie-Hellman of WRITER,READER)
AES: CTR mode (starting 0, nonce 0) using AESKEY of DATA
DATA: gzip([TITLE\\0CONTENTS\\0]+), padded with 0 bytes to 8054\0"""

FULL_LENGTH = 8192
DATA_LENGTH = FULL_LENGTH - (64 + 33 + 33 + 8)
RECORD_LENGTH = len(preamble) + FULL_LENGTH
# BIP340 tag excludes final \0
bip340tag = verheader[:-1]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['centurymetadata', 'centurymetadata.server']

package_data = \
{'': ['*']}

install_requires = \
['pycryptodomex>=3.6', 'secp256k1>=0.14.0']

setup_kwargs = {
    'name': 'centurymetadata',
    'version': '0.2',
    'description': 'Encrypt, send, retrieve and decrypt centurymetadata.org files',
    'long_description': '# centurymetadata.org: Long-term Bitcoin Metadata Storage\n\n## About\n\n*Century Metadata* is a project to provide storage for small amounts\nof auxiliary data.  As an example, this is useful for Bitcoin wallets,\nwhich can be restored from 12 seed words, but cannot know about more\ncomplex funds without additional data.  On restore, your wallet would attempt to\nfetch this data from [https://centurymetadata.org](https://centurymetadata.org) or a mirror.\n\nWe are currently in alpha, seeking feedback.\n\n## File Format\n\nThe file format is designed to be self-explanatory and use standard,\nlong-lived primitives as much as possible.  Every file contains a\npreamble, followed by 8192 bytes.  The preamble describes the data\nformat which follows:\n\n```\ncenturymetadata v0\\0SIG[64]|WRITER[33]|READER[33]|GEN[8]|AES[8054]\n\nSIG: BIP-340 SHA256(TAG|TAG|WRITER|READER|GEN|AES)\nWRITER, READER: secp256k1 x-only keys\nTAG: SHA256("centurymetadata v0"[18])\nAESKEY: SHA256(EC Diffie-Hellman of WRITER,READER)\nAES: CTR mode (starting 0, nonce 0) using AESKEY of DATA\nDATA: gzip([TITLE\\0CONTENTS\\0]+), padded with 0 bytes to 8054\\0\n```\n\nThe data itself is a series of NUL-separated title, contents pairs.\nObviously this cannot be validated on the production server, but the\ntest server (which only allows known keys) will check the file is\ncompliant.\n\n## Usage with Bitcoin\n\nThe BIP 32 path recommended for centurymetadata is `0x44315441\'`\n(`DATA`), with `/0\'` as the writer key,\n`/1\'` as the reader key.  Of course, others can also send data\nto your reader key, but you know that the record from your own writer\nkey can be trusted. \n\nThe types of records accepted are as follows:\n\n* Title: `bitcoin psbt`, Body: base64-encoded PSBT\n* Title: `bitcoin transaction` Body: hex-encoded transaction\n* Title: `bitcoin miniscript` Body: miniscript string\n\n## API\n\nThe test API endpoint can be found at [testapi.centurymetadata.org](https://testapi.centurymetadata.org/api/v0).\n\n### Entry Creation: POST /api/v0/authorize/{READER}/{WRITER}/{AUTHTOKEN}\n\nYou need to get an *AUTHTOKEN* for each new entry.  There can only be\none entry for any *READER*/*WRITER* pair, but once the entry is\nauthorized it can be updated by the writer at any time.\n\n### Entry Update: POST /api/v0/update\n\nUpdates a previously authorized writer/reader entry.  The\n`Content-Type: application/x-centurymetadata` should contain a valid\ncenturymetadata file.\n\n### Entries Depth: GET /api/v0/fetchdepth\n\nSince we bundle records by reader prefix (e.g. all readers starting with `42a3` might be bundled together), you need to know how long the prefix is: it starts as an empty prefix and increases by one hex digit as we grow, so bundles are always a reasonable size.\n\nReturns a JSON object with member `depth` containing how many hex digits of reader to use for `fetchbundle`.\n\n### Retrieiving Entries: GET /api/v0/fetchbundle/{READERPREFIX}\n\nThis returns the given bundle, as `Content-Type: application/x-centurymetadata`, consisting of multiple back-to-back\ncentury metadata files.\n\n## Tools\n\nThere is an experimental Python package to encode and decode\ncenturymetadata files in the [GitHub repository](https://github.com/rustyrussell/centurymetadata)\n\n## Roadmap\n\nI\'m committed to maintaining this service for at least 5 years\nas a trial.  After that if it\'s proven useful I would like to\nspin it into a real not-for-profit foundation to provide as much\ncertainty on continuity as possible.\n\n## How Much?\n\nThere will never be a charge for ratelimited updates or retrievals;\nthe idea is to charge a small cost for the creation of new entries to\ncover ongoing running costs.  We may also accept donations.\n\n## Who?\n\nRusty Russell started this as a side project; my original problem was\nhow to give someone timelocked bitcoin, but realized there was a large\nrelated class of problems for someone to solve.\n\n## Feedback\n\nAdvice, suggestions, kudos, blame: hosting is on [GitHub](https://github.com/rustyrussell/centurymetadata), and you can reach us on [Twitter](https://twitter.com/centurymetadata), or send\n[me email](mailto:rusty@rustcorp.com.au) or other contact as listed on \n[my personal site](https://rusty.ozlabs.org).\n',
    'author': 'Rusty Russell',
    'author_email': 'rusty@rustcorp.com.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://centurymetadata.org/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



#! /usr/bin/python3
import cgi
import os
import json
import sys
import centurymetadata
from secp256k1 import PublicKey
from typing import Optional, Any

TOPLEVEL = "/api/v0/"
BASEDIR = "/home/rusty/data/centurymetadata/"


def bad_404() -> None:
    print('Status: 404\n\n<html><head></head><body>Invalid URL, see <a href="https://github.com/rustyrussell/centurymetadata/tree/master/examples/EXAMPLES.md">EXAMPLES.md</a></body></html>')
    exit(0)


def bad_405() -> None:
    print("Status: 405\n\nMethod Not Allowed")
    exit(0)


def bad_400(extra: str) -> None:
    print("Status: 400\n\nBad Request ({})".format(extra))
    exit(0)


def bad_403(extra: str) -> None:
    print("Status: 403\n\nForbidden ({})".format(extra))
    exit(0)


def success(msg: str = 'Success', ctype: str = 'text/plain') -> None:
    print("Content-Type: {}\n".format(ctype))
    print(msg)
    exit(0)


def decode_key(hexkey: str) -> Optional[PublicKey]:
    try:
        k = PublicKey(bytes.fromhex(hexkey), raw=True)
    except ValueError:
        return None
    return k


def storage_dir(rkey: PublicKey, wkey: PublicKey) -> str:
    return os.path.join(BASEDIR, rkey.serialize().hex() + "+" + wkey.serialize().hex())


def authorize(reader: str, writer: str, authtoken: str) -> None:
    # We use a dummy authtoken for testapi
    if authtoken != '0' * 64:
        return bad_403("AUTHTOKEN must be all-zero for testapi")
    wkey = decode_key(writer)
    rkey = decode_key(reader)
    if wkey is None:
        return bad_400("writer must be valid compressed pubkey")
    if rkey is None:
        return bad_400("reader must be valid compressed pubkey")

    try:
        os.mkdir(storage_dir(rkey, wkey))
    except FileExistsError:
        bad_400("READER {} WRITER {} already authorized"
                .format(writer, reader))
    success()


def update() -> None:
    content = os.getenv("CONTENT_TYPE")
    if content != 'application/x-centurymetadata':
        return bad_400("update must be Content-Type: application/x-centurymetadata")
    bytelen = int(os.getenv("CONTENT_LENGTH") or 0)
    b = bytes()
    while bytelen > 0:
        r = sys.stdin.buffer.read(bytelen)
        bytelen -= len(r)
        b += r
    wkey, rkey, gen, after_pre = centurymetadata.deconstruct(b)

    if not centurymetadata.check_sig(after_pre):
        return bad_400("Bad signature on x-centurymetadata")

    # OK, signature checks out.
    try:
        f = open(os.path.join(storage_dir(rkey, wkey), gen.to_bytes(8, "big").hex()), "xb")
    except FileExistsError:
        return bad_400("Generation {} already exists".format(gen))
    except FileNotFoundError:
        return bad_403("Writer {} reader {} not authorized; try authorize?"
                       .format(wkey.serialize().hex(), rkey.serialize().hex()))

    f.write(b)
    f.close()
    success()


def fetchdepth() -> None:
    """Get the length of prefix we're supposed to use for fetchbundle"""
    with open(os.path.join(BASEDIR, "maxdepth"), "r") as depthf:
        maxdepth = int(depthf.read())

    success(ctype='application/json',
            msg='{"depth":' + str(maxdepth) + '}')


def fetchbundle(prefix: str) -> None:
    if any(c not in "0123456789abcdef" for c in prefix):
        return bad_400("prefix must be lowercase hex characters!")

    with open(os.path.join(BASEDIR, "maxdepth"), "r") as depthf:
        maxdepth = int(depthf.read())
    with open(os.path.join(BASEDIR, "mindepth"), "r") as depthf:
        mindepth = int(depthf.read())

    # For transitions, we allow a range, but really we expect maxdepth.
    if len(prefix) < mindepth or len(prefix) > maxdepth:
        return bad_400("prefix must be {} characters: see fetchdepth!"
                       .format(maxdepth))

    print("Content-Type: application/x-centurymetadata\n")
    sys.stdout.flush()
    entries = sorted([d for d in os.listdir(BASEDIR) if d.startswith(prefix)])

    for f in entries:
        gens = sorted(os.listdir(os.path.join(BASEDIR, f)))
        # Might be authorized, but never updated, otherwise send last.
        if len(gens) != 0:
            with open(os.path.join(BASEDIR, f, gens[-1]), "rb") as serv:
                sys.stdout.buffer.write(serv.read())

    sys.stdout.buffer.flush()


requests = {'authorize': ("POST", authorize, 3),
            'update': ("POST", update, 0),
            'fetchdepth': ("GET", fetchdepth, 0),
            'fetchbundle': ("GET", fetchbundle, 1)}

req = os.getenv("PATH_INFO")
reqmethod = os.getenv("REQUEST_METHOD")
if not req or not reqmethod:
    cgi.test()
    exit(1)

if not req.startswith(TOPLEVEL):
    bad_404()

reqparts = req.split('/')[3:]
if reqparts[0] not in requests:
    bad_404()

handler: Any
method, handler, numargs = requests[reqparts[0]]
if reqmethod != method:
    bad_405()

# In case last one is empty (i.e. ends in /):
if reqparts[-1] == '':
    reqparts = reqparts[0:-1]

if len(reqparts[1:]) != numargs:
    bad_400("Expected {} args, got {}".format(numargs, len(reqparts[1:])))

handler(*reqparts[1:])
